/****************************************************************************
*   PROJECT: Squeak port for Win32 CE (NT / Win95)
*   FILE:    sqWinCEAlloc.c
*   CONTENT: Time routines is not included in the WCE Memory management.
*
*   AUTHOR:  Alejandro F. Reimondo
*   ADDRESS: Sugar - http://www.sugarWeb.com
*   EMAIL:   aleReimondo@sugarWeb.com
*   RCSID:   
*
*   NOTES:
*
*****************************************************************************/
#include <windows.h>
#include "sq.h"

extern void exit(int status);
void deleteBackingStoreFile();

static int useBackingStoreFlag = 0;

#if 0
// Reimplement malloc in order to catch out of memory conditions once. This saves
// repeating the code, since there is nowt we can do to recover anyway.
void* malloc(size_t size)
{
	void* pBlock = LocalAlloc(LPTR, size);
	if (pBlock == NULL)
	{
		SHShowOutOfMemory(0,0);
		exit(-1);
	}
	return pBlock;
}

void free(void* memblock)
{
	LocalFree(memblock);
}

#endif

void *calloc(size_t num, size_t size)
{
	void *ptr = malloc(num * size);
	memset(ptr, 0, num*size);
	return ptr;
}

#if defined(NO_VIRTUAL_MEMORY)
// Rather than malloc() enormous block from the heap, grab it straight from lowest
// level allocator. We have to do this for CE anyway because the local heap cannot
// exceed 1Mb in size
#undef sqAllocateMemory
#ifndef USE_BACKINGSTORE
void* sqAllocateMemory(int minHeapSize, int desiredHeapSize)
{
        return VirtualAlloc(NULL, desiredHeapSize, MEM_COMMIT, PAGE_READWRITE);
}
#else
void * sqAllocateMemory(int minHeapSize, int desiredHeapSize)
{
  void * pMappedMem = VirtualAlloc(NULL, desiredHeapSize, MEM_COMMIT, PAGE_READWRITE);
  if (!useBackingStoreFlag) {
    return pMappedMem;
  }
  
  if (!pMappedMem) {
    HANDLE hFile = CreateFileForMapping(BACKINGSTOREFILE, GENERIC_READ | GENERIC_WRITE,
					FILE_SHARE_WRITE, NULL, CREATE_ALWAYS,
					FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS, 0);

    if (hFile != INVALID_HANDLE_VALUE) {
      HANDLE hMapping = CreateFileMapping(hFile, NULL, PAGE_READWRITE, 0,
					  desiredHeapSize, 0);
      if (hMapping) {
	pMappedMem = (unsigned char *) MapViewOfFile(hMapping, FILE_MAP_ALL_ACCESS, 0, 0, 0);
      }
    }
  }

  return pMappedMem;
}

void deleteBackingStoreFile()
{
  if (DeleteFile(BACKINGSTOREFILE) == 0) {
    ;
  }
}

#endif
#endif // defined(NO_VIRTUAL_MEMORY)

// Initialize the object memory space
int initializeCEMemorySpace(int imageSize)
{	
	DWORD memSize = (DWORD)-1;

	// Determine the amount of available memory
	MEMORYSTATUS memStats;
	memStats.dwLength = sizeof(memStats);
	GlobalMemoryStatus(&memStats);
	
	// The max. mem. size should leave some free space for other apps
	// and the system
	memSize = memStats.dwAvailPhys - 128*1024;
	
	memSize = min(memSize, (12*1024*1024));
#ifdef USE_BACKINGSTORE
	{
	  ULARGE_INTEGER availableToCaller;
	  ULARGE_INTEGER totalNumberOfBytes;
	  ULARGE_INTEGER totalNumberOfFreeBytes;
	  DWORD backingStoreMemSize = (DWORD) -1;

	  deleteBackingStoreFile();

	  if (imageSize < memSize - (2*1024*1024)) {
	    return memSize;
	  }
	  
	  if (GetDiskFreeSpaceEx(BACKINGSTOREPATH,
				 &availableToCaller, 
				 &totalNumberOfBytes,
				 &totalNumberOfFreeBytes) == 0) {
	    return memSize;
	  }

	  if (availableToCaller.HighPart != 0) {
	    backingStoreMemSize = 500*1024*1024;
	  } else {
	    backingStoreMemSize = availableToCaller.LowPart - (1*1024*1024);
	  }
	  if (memSize < backingStoreMemSize) {
	    useBackingStoreFlag = 1;
	    memSize = backingStoreMemSize;
	  }
	}
#endif	
	return memSize;
}
